#include <DebugOverSoftSerial.h>
#include <SimpleSim800.h>
#include <LowPower.h>

#define VERSION "&v=\"0.23V\"&key=\"0815\""

#define LED_RED_PIN          8
#define LED_GREEN_PIN        9
#define LED_BLUE_PIN        13
#define LED_DELAY_BLINK_MS  65

#define SWITCH_ONE_PIN       2
#define SWITCH_TWO_PIN       3

#define KEEP_VOLTAGE_PIN     6

#define BURN_READING_COUNT   8

// ~5 minutes / 8 seconds
// duration of loop ist a little bit longer than 8 seconds
// so we chose a 10% lower value
#define INITIAL_TIME_CHECK_CYCLE_FACTOR   35
// ~6 hours / 8 seconds
// duration of loop ist a little bit longer than 8 seconds
// so we chose a 10% lower value
#define LONG_TIME_CHECK_CYCLE_FACTOR    2350
// ~1 hour / 8 seconds
#define TIME_CHECK_CYCLE_FACTOR_HOUR     450

#define SLEEP_PERIOD  SLEEP_8S

const int BATTERY_SENSE_PIN = A3;  // select the input pin for the battery sense point
/*
 * Calcolation for Voltage divider
 * low resistor = 2.2kOhm
 * high resistor = 10kOhm
 * reference voltage = 1.1V
 * claculation
 * Vpin = Vmeasure * 2.2 / (2.2 + 10)
 * Vpin = 1.1V * Xadc / 1023
 * Vmeasure = (1.1V / 1023) * ((2.2 + 10) / 2.2) * Xadc
 * Vmeasure = 0.005963V * Xadc
 */
// const float VOLTAGE_CALCULAION_FACTOR = 5.963;
// try avoiding float calculation
const byte VOLTAGE_CALCULAION_FACTOR =  61;
const byte VOLTAGE_CALCULAION_DIVISOR = 10;
const int BROWN_OUT_VOLTAGE = 3200;

const byte MESSAGE_TIMES[] = {0,6,12,18};

DebugOverSoftSerial logger;
// APN and SIM are configured directly in library
SimpleSim800 sim800;

const char m_Url[] = "http://example.com/traps/ping";
byte m_NextMessageTimeIndex = 0;
int m_TimeCheckCycleFactor = 0;
byte m_ActualHour = 0;

char m_Imei[20] = {0};
char m_VoltageString[10] = {0};
unsigned int m_VoltageValue = 0;
char m_SignalStrength[10] = {0};
char m_Iccid[22] = {0};
char m_PostData[100] = {0};

bool m_LastSwitchOneValue = 0;
bool m_LastSwitchTwoValue = 0;

int m_LoopCounter = 0;
byte m_LowVoltageCounter = 0;
bool m_IsFirstLoop = true;

int m_year, m_month, m_day, m_hour, m_minute, m_second;
char m_dateString[22] = {0};

void setup() 
{
  char txBufferSizeString[4] = {0};
  itoa(SERIAL_TX_BUFFER_SIZE, txBufferSizeString, 10);
  logger.print("Serial TX Buffer is ");
  logger.print(txBufferSizeString);
  logger.println(" bytes");
    
  pinMode(KEEP_VOLTAGE_PIN, OUTPUT);
  digitalWrite(KEEP_VOLTAGE_PIN, HIGH);
  
  // set some ports
  pinMode(SWITCH_ONE_PIN, INPUT);
  pinMode(SWITCH_TWO_PIN, INPUT);

  pinMode(LED_RED_PIN, OUTPUT);
  pinMode(LED_GREEN_PIN, OUTPUT);
  pinMode(LED_BLUE_PIN, OUTPUT);

  // Ports are set
  blinkGreen(1);
  
  // set reference for voltage measurement
  analogReference(INTERNAL);

  // start GSM module
  sim800.begin();
  // GSM module started
  blinkGreen(2);
  // this delay is necessary, it helps the device to be ready and connect to a network
  delay(5000); 

  initializeGprs();   
  sim800.getImei(m_Imei);
  sim800.getIccid(m_Iccid);

  hourWithTimezone(false);
  m_ActualHour = m_hour;
  for (byte i = 0; i < sizeof(MESSAGE_TIMES); i++)
  {
    int messageTimeAtIndex = MESSAGE_TIMES[i];
    if (messageTimeAtIndex > m_ActualHour)
    {
      m_NextMessageTimeIndex = i;
      break;
    }
  }

  // lookup of time will happen about every 5 minutes
  m_TimeCheckCycleFactor = INITIAL_TIME_CHECK_CYCLE_FACTOR;
}

void loop() 
{
  bool execReportToServer = false;
  
  // kind of heartbeat
  blinkGreen(1);

  // read the volage
  readVoltage();

  // read the switches
  bool actualSwitchOneValue = digitalRead(SWITCH_ONE_PIN);
  bool actualSwitchTwoValue = digitalRead(SWITCH_TWO_PIN);

  // If this ist the first loop, report to server
  if(m_IsFirstLoop)
  {
      execReportToServer = true;
  }
  
  // do we need to report becaus of switch changes?
  if(actualSwitchOneValue != m_LastSwitchOneValue ||
     actualSwitchTwoValue != m_LastSwitchTwoValue)
  {
    execReportToServer = true;
  }

  // if voltage is low
  if (m_VoltageValue < BROWN_OUT_VOLTAGE && m_LowVoltageCounter > 3)
  {
    logger.println("Shutting down ...");
    delay(5000);
    digitalWrite(KEEP_VOLTAGE_PIN, LOW);
  }
  else if (m_VoltageValue < BROWN_OUT_VOLTAGE && m_LowVoltageCounter > 2)
  {
    execReportToServer = true;
    m_LowVoltageCounter++;
  }
  else if (m_VoltageValue < 3000)
  {
    m_LowVoltageCounter++;
  }
  else
  {
    m_LowVoltageCounter = 0;
  }

  // we don't need to check the time every cycle
  // assume a time of 8s per cycle
  if (m_LoopCounter >= m_TimeCheckCycleFactor)
  {
    // We have three TimeCheckCycleFactor
    //
    // ~5 minutes / 8 seconds
    // INITIAL_TIME_CHECK_CYCLE_FACTOR -   35
    //
    // ~1 hour / 8 seconds
    // TIME_CHECK_CYCLE_FACTOR_HOUR    -  450
    //
    // ~6 hours / 8 seconds
    // LONG_TIME_CHECK_CYCLE_FACTOR    - 2430
    //
    // We have two hour-Values
    // m_ActualHour       - the actual hour got by SIM800
    // messageTimeAtIndex - the next full hour to report
    //
    // first we get the actualHour
    hourWithTimezone(true);
    if(m_year < 2020)
    {
      initializeGprs();
      hourWithTimezone(true);
    }
    m_ActualHour = m_hour;

    // we may have to correct the actualHour
    if (m_ActualHour < 0 || m_ActualHour >= 24)
    {
        m_ActualHour = 0;
    }

    // we log the actualHour
    char actualHourAsString[3] = {0};
    itoa(m_ActualHour, actualHourAsString, 10);
    logger.print("Actual hour is ");
    logger.println(actualHourAsString);

    // default is that we are more than one hour away from reporting-time
    m_TimeCheckCycleFactor = TIME_CHECK_CYCLE_FACTOR_HOUR; 

    // we calculate the next hour for reporting
    int messageTimeAtIndex = MESSAGE_TIMES[m_NextMessageTimeIndex];
    char messageTimeAsString[3] = {0};;
    itoa(messageTimeAtIndex, messageTimeAsString, 10);
    logger.print("Hour at index is ");
    logger.println(messageTimeAsString);
      
    //
    // next we have to compare the two hour-Values
    //
    if ((m_ActualHour == messageTimeAtIndex) ||
        (m_NextMessageTimeIndex > 0 && messageTimeAtIndex < m_ActualHour))
    {
      // two szenarios, nearly the same doing
      //
      // szenario 1:
      // we are on the correct hour to report
      //
      // szenario 2:
      // we are over report hour, maybe we have slept too long
      //
  
      // first we increment the index to the next message-time
      m_NextMessageTimeIndex++;
      if (m_NextMessageTimeIndex >= sizeof(MESSAGE_TIMES))
      {
        m_NextMessageTimeIndex = 0;
      }

      // we report to server
      execReportToServer = true;

      // we can set the timeCheckCycleFactor to a value
      // that the lookup will happen just about before the next hour
      if(m_IsFirstLoop || messageTimeAtIndex < m_ActualHour)
      {
        // szenario 1 while startup or szenario 2
        m_TimeCheckCycleFactor = TIME_CHECK_CYCLE_FACTOR_HOUR;
      }
      else
      {
        // regular szenario 1
        m_TimeCheckCycleFactor = LONG_TIME_CHECK_CYCLE_FACTOR;
      }
    }
    else if ((messageTimeAtIndex == 0 && m_ActualHour == 23) ||
             (messageTimeAtIndex > 0 && m_ActualHour + 1 == messageTimeAtIndex))
    {
        // we are in the hout before reporting-time
        // we cycle only in 5 minute steps
        m_TimeCheckCycleFactor = INITIAL_TIME_CHECK_CYCLE_FACTOR;
    }
      
    m_LoopCounter = 0;
  }

  if (execReportToServer)
  {
    if(reportToServer(actualSwitchOneValue, actualSwitchTwoValue, m_IsFirstLoop) == 0)
    {
        m_LastSwitchOneValue = actualSwitchOneValue;
        m_LastSwitchTwoValue = actualSwitchTwoValue;
    }
  }
  
  m_IsFirstLoop = false;

  m_LoopCounter++;
  LowPower.powerDown(SLEEP_PERIOD, ADC_OFF, BOD_ON);
}

void initializeGprs()
{
  if(sim800.init() != 0) {
      delay(1000);
      blinkRed(2);
  }

  if(sim800.connect() != 0) {
      delay(1000);
      blinkRed(3);
  }
  delay(3000); 
  blinkGreen(3);
}

int reportToServer(byte switchOne, byte switchTwo, bool submitIccid)
{
  digitalWrite(LED_BLUE_PIN, HIGH);
  memset(m_VoltageString, 0, sizeof(m_VoltageString));
  itoa(m_VoltageValue, m_VoltageString, 10);
  
  memset(m_PostData, 0, sizeof(m_PostData));
  // append IMEI
  strcat(m_PostData, "imei=");
  strcat(m_PostData, m_Imei);

  if (submitIccid)
  {
    strcat(m_PostData, "&iccid=");
    strcat(m_PostData, m_Iccid);  
  }
  
  // append voltage
  strcat(m_PostData, "&mv=");
  strcat(m_PostData, m_VoltageString);

      // append switch-values
  if (switchOne == 0)
  {
    strcat(m_PostData, "&s1=0");
  }
  else
  {
    strcat(m_PostData, "&s1=1");
  }
    
  if (switchTwo == 0)
  {
    strcat(m_PostData, "&s2=0");
  }
  else
  {
    strcat(m_PostData, "&s2=1");
  }
  strcat(m_PostData, VERSION);

  // wake up GSM-module
  if(sim800.connect() == 0)
  {
    memset(m_SignalStrength, 0, sizeof(m_SignalStrength));
    sim800.getSignalStrength(m_SignalStrength);
    strcat(m_PostData, "&strength=");
    strcat(m_PostData, m_SignalStrength);

    byte postLoopCounter = 0;
    byte result = sim800.postData(m_Url, m_PostData);
    while (result != 0 && postLoopCounter < 2)
    {
      postLoopCounter++;
      result = sim800.postData(m_Url, m_PostData);
    }
    
    sim800.disconnect();
    digitalWrite(LED_BLUE_PIN, LOW);
    return 0;
  }
  digitalWrite(LED_BLUE_PIN, LOW);
  blinkRed(10);
  return 1;
}

void blinkGreen(byte amount)
{
  for (byte i = 0; i < amount; i++)
  {
    blink(LED_GREEN_PIN, LED_DELAY_BLINK_MS);
  }
}

void blinkRed(byte amount)
{
  for (byte i = 0; i < amount; i++)
  {
    blink(LED_RED_PIN, LED_DELAY_BLINK_MS);
  }
}

void blink(byte pin, byte delayTime)
{
  pinMode(pin, OUTPUT);
  digitalWrite(pin,HIGH);
  delay(delayTime);
  digitalWrite(pin,LOW);
  delay(delayTime);
}

void hourWithTimezone(bool withPowerSaving)
{
  if (withPowerSaving)
  {
    sim800.disablePowerSaving();
  }
  
  // Answer of SIM800L looks like
  // yy/MM/dd,hh:mm:ss+/-zz
  memset(m_dateString, 0, sizeof(m_dateString));
  byte counter, offset;
  
  m_year = 2000;
  counter = 0;

  do 
  {
    if (sim800.getDateTime(m_dateString) == 0)
    {
      logger.print("Actual date is ");
      logger.println(m_dateString);
    
      sscanf(m_dateString, "%d/%d/%d,%d:%d:%d+00", &m_year, &m_month, &m_day, &m_hour, &m_minute, &m_second);
      m_year = m_year + 2000;
      // Starting with MET
      offset = 1;
      if(sim800.isSummertime(m_year, m_month, m_day, m_hour, offset))
      {
        // Seems we've got MEST
        offset = 2;
      }
      m_hour =  m_hour + offset;
      if (m_hour > 23)
      {
        m_hour = m_hour - 24;
      }
    }
    else
    {
      m_hour = 255;
    }
    counter++;
    if (m_year < 2021)
    {
      delay(10000);
    }
  }
  while (m_year < 2021 && counter < 3);
  
  if (withPowerSaving)
  {
    sim800.enablePowerSaving();
  }
}

void burnReadings(int pin, int counter)
{
  for (int i = 0; i < counter; i++)
  {
    analogRead(pin);
  }
}

/**
 * returns volatge in mV
 */
static void readVoltage()
{
  analogReference(INTERNAL);
  burnReadings(BATTERY_SENSE_PIN, BURN_READING_COUNT);
  delay(10); 
  int result = analogRead(BATTERY_SENSE_PIN);
  unsigned int voltage = VOLTAGE_CALCULAION_FACTOR * result;
  voltage = voltage / VOLTAGE_CALCULAION_DIVISOR;
  m_VoltageValue = voltage;
}